import { Injectable, Logger } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { ChickensService } from 'src/chickens/chickens.service';
import { CODE } from 'src/common/code';
import { GoodsStatus, GoodsTypes, UserOnlieStatus } from 'src/common/const';
import { ExternalApiService } from 'src/common/utils/external-api';
import { CustomResult, DataResultModel, SuccessResult } from 'src/common/utils/result-utils';
import { GoodsService } from 'src/goods/goods.service';
import { CreateOrderlogDto } from 'src/orderlog/dto/create-orderlog.dto';
import { OrderlogService } from 'src/orderlog/orderlog.service';
import { getUserShoppingKey, getUserTokenKey } from 'src/redis/keys';
import { CacheService } from 'src/redis/redis';
import { getManager, Repository } from 'typeorm';
import { BuyGoodsUserDto } from './dto/buygoods-user.dto';
import { CreateUserDto } from './dto/create-user.dto';
import { UpdateUserDto } from './dto/update-user.dto';
import { UserEntity } from './entities/user.entity';
import { UserModel, AppUserModel } from './user.model';

const prefix = 'user';
const logger = new Logger('UserService');
@Injectable()
export class UserService {
  constructor(
    @InjectRepository(UserEntity)
    private readonly userRepository: Repository<UserEntity>,
    private readonly goodsService: GoodsService,
    private readonly orderLogService: OrderlogService,
    private readonly cacheService: CacheService,
    private readonly chickensService: ChickensService,
  ) { }

  /**查询用户信息,没有则初始化 */
  async initQuery(createUserDto: CreateUserDto): Promise<DataResultModel> {
    //初始化token
    await this.initToken(createUserDto);
    //查询用户信息
    const userInfo = await this.findOne(createUserDto.userid);
    if (userInfo) {
      //小鸡信息
      userInfo['chickens'] = {};
      const chick = await this.chickensService.chickensInfo(createUserDto.userid);
      if (chick) userInfo['chickens'] = chick;
      return SuccessResult(userInfo);
    }
    //有信息则返回
    const appUserInfo = await ExternalApiService.getAppUserInfo(createUserDto.userid, createUserDto.token);
    //用户信息异常处理
    if (!appUserInfo || appUserInfo.code != 200) return appUserInfo;
    //没有信息则初始化
    const initRes = await this.initUser(appUserInfo.data);
    if (!initRes) return CustomResult(CODE.CODE_INIT_USER_ERR);

    return SuccessResult(initRes);
  }
  /**用户信息初始化 */
  private async initUser(appUser: AppUserModel) {
    const entity: UserModel = {
      userid: appUser.id,
      avatar: appUser.avatar,
      nickname: appUser.nickname,
      fodder_balance: 0,
      eggs: 0,
      online_status: UserOnlieStatus.ONLINE,
      parent_id: 0,
      create_time: new Date(),
      update_time: new Date()
    }
    return await this.userRepository.save(entity);
  }


  /**查询用户信息 */
  async findOne(userid: number) {
    const info = await this.userRepository.createQueryBuilder(prefix)
      .select()
      .where('userid = :userid', { userid: userid })
      .getRawOne();
    return info;
  }

  /**用户token初始化 */
  async initToken(createUserDto: CreateUserDto) {
    const tokenKey = getUserTokenKey(createUserDto.userid);
    //查询Redis token 信息
    const tokenInfo = await this.cacheService.get(tokenKey);
    if (tokenInfo) return SuccessResult(tokenInfo);
    //存储Redis
    await this.cacheService.set(tokenKey, createUserDto.token);
    return SuccessResult(createUserDto.token);
  }

  /**更新用户数据 */
  async update(userid: number, updateUserDto: UpdateUserDto) {
    updateUserDto.update_time = new Date();
    const res = await this.userRepository.createQueryBuilder()
      .update()
      .set(updateUserDto)
      .where('userid = :userid', { userid: userid })
      .execute();
    if (!res) return CustomResult(CODE.CODE_USER_UPDATE_ERR)
    return SuccessResult();
  }

  /** 用户购买商品逻辑 */
  async buy(buyGoodsUserDto: BuyGoodsUserDto) {
    const { userid, goodsid, goods_count } = buyGoodsUserDto;
    //查询APP用户信息
    const token = await this.cacheService.get(getUserTokenKey(userid));
    const appUserInfo = await ExternalApiService.getAppUserInfo(userid, token);
    if (appUserInfo?.code != 200) return appUserInfo;
    const user = appUserInfo.data;

    //查询商品信息
    const goodsInfo: any = await this.goodsService.findOne(goodsid, [GoodsStatus.SELLING]);
    if (goodsInfo?.code != 200) return CustomResult(CODE.CODE_GOODS_INFO_ERR);
    const goods = goodsInfo.data;

    //判断用户金额是否满足条件 (商品价格*商品数量)
    const payCoin = goods.goods_price * goods_count;//应支付价格
    if (user.coin < payCoin) return CustomResult(CODE.CODE_APPUSER_COIN_NOT_ENOUGHT_ERR);

    //扣减加锁 锁几秒
    const locKeys = getUserShoppingKey(userid, goodsid);
    const queryLock = await this.cacheService.get(locKeys);
    if (queryLock) return CustomResult(CODE.CODE_USER_PAY_LOCK_ERR);
    await this.cacheService.set(locKeys, true, 3);

    //用户金额扣减
    const subUserInfo = await ExternalApiService.subAppUserCoin(userid, token, payCoin);

    if (subUserInfo?.code != 200) return CustomResult(CODE.CODE_APPUSER_SUB_COIN_ERR);
    // 购买饲料则要发放饲料
    if (goods.goods_type === GoodsTypes.FODDER) {
      await this.insertFodder(userid, goods.goods_attribute * goods_count);
    }

    //发放完成记录日志
    const orderInfo: CreateOrderlogDto = {
      userid: buyGoodsUserDto.userid,
      goods_id: goods.goods_id,
      goods_type: goods.goods_type,
      goods_price: goods.goods_price,
      goods_desc: goods.goods_desc || '',
      goods_url: goods.goods_image || '',
      goods_attribute: goods.goods_attribute
    }
    const orderRes = await this.orderLogService.create(orderInfo);
    logger.debug(`购买订单日志结果: ${JSON.stringify(orderRes)}`);
    if (orderRes.code != 200) return CustomResult(CODE.CODE_GOODS_DEAL_ERR);

    //返回用户信息
    const userInfo = await this.userRepository.createQueryBuilder(prefix).select().where('userid=:userid', { userid: userid }).getRawOne();
    return SuccessResult(userInfo);
  }

  /**
   * 饲料领取
   */
  private async insertFodder(userid: number, fodder: number) {
    await getManager().transaction(async manager => {
      const user = await manager.findOne(UserEntity, { userid: userid });
      if (user) {
        user.update_time = new Date();
        user.fodder_balance = Number(user.fodder_balance) + fodder;
        await manager.save(user);
        logger.debug(`用户 ${userid} 购买了 ${fodder} g饲料`);
      }
    })
  }
}
